from typing import Type

from bottle import request
from injector import inject

from antibot.internal.backend.debugger import Debugger
from antibot.internal.backend.endpoint_runner import EndpointRunner
from antibot.internal.backend.request_checker import RequestChecker
from antibot.internal.slack.channel import Channel
from antibot.plugin import AntibotPlugin
from antibot.repository.users import UsersRepository
from antibot.slack.api import SlackApi
from antibot.slack.message import Message


class CommandRunner:
    @inject
    def __init__(self, endpoints: EndpointRunner, users: UsersRepository, checker: RequestChecker,
                 api: SlackApi, debugger: Debugger):
        self.endpoints = endpoints
        self.users = users
        self.checker = checker
        self.api = api
        self.debugger = debugger

    def run_command(self, method, plugin: Type[AntibotPlugin]):
        self.checker.check_request(request)

        user = self.users.get_user(request.forms['user_id'])
        channel = Channel(request.forms['channel_id'], request.forms['channel_name'])
        response_url = request.forms['response_url']
        params = request.forms['text']

        with self.debugger.wrap({'params': params}):
            reply = self.endpoints.run(plugin, method, user=user, channel=channel, response_url=response_url,
                                       params=params)

        if isinstance(reply, Message):
            return self.api.respond(response_url, reply)



import json

from bottle import request
from injector import singleton, inject

from antibot.internal.backend.actions.block_action_runner import BlockActionRunner
from antibot.internal.backend.actions.view_closed_runner import ViewClosedRunner
from antibot.internal.backend.actions.view_submit_runner import ViewSubmitRunner
from antibot.internal.backend.debugger import Debugger
from antibot.internal.backend.request_checker import RequestChecker
from antibot.repository.users import UsersRepository


@singleton
class ActionRunner:
    @inject
    def __init__(self, users: UsersRepository, checker: RequestChecker, block_actions: BlockActionRunner,
                 debugger: Debugger, view_closed: ViewClosedRunner, view_sumbit: ViewSubmitRunner):
        self.users = users
        self.block_actions = block_actions
        self.checker = checker
        self.debugger = debugger
        self.view_closed = view_closed
        self.view_sumbit = view_sumbit

    def run(self):
        self.checker.check_request(request)

        json_data = json.loads(request.forms['payload'])

        with self.debugger.wrap(json_data):
            if json_data['type'] == 'block_actions':
                return self.block_actions.run_callback(json_data)
            elif json_data['type'] == 'view_closed':
                return self.view_closed.run(json_data)
            elif json_data['type'] == 'view_submission':
                return self.view_sumbit.run(json_data)



import re

from pyckson import no_camel_case


@no_camel_case
class File:
    def __init__(self, id: str, permalink: str, permalink_public: str, name: str):
        self.id = id
        self.permalink = permalink
        self.permalink_public = permalink_public
        self.name = name

        match = re.match(r'https://slack-files.com/(.*)-(.*)-(.*)', self.permalink_public)
        self.team_id = match.group(1)
        self.secret = match.group(3)

    @property
    def public_link(self):
        return 'https://files.slack.com/files-pri/{}-{}/{}?pub_secret={}'.format(
            self.team_id,
            self.id,
            self.permalink.split('/')[-1],
            self.secret
        )



from typing import Iterable, Type

import pkg_resources
from injector import Module

from antibot.plugin import AntibotPlugin


class PluginsCollection:
    def __init__(self, plugins):
        self.plugins = plugins

    def __iter__(self):
        return iter(self.plugins)


def find_plugins() -> Iterable[Type[AntibotPlugin]]:
    for entry_point in pkg_resources.iter_entry_points('antibot'):
        object = entry_point.load()
        if issubclass(object, AntibotPlugin):
            yield object


def find_modules() -> Iterable[Type[Module]]:
    for entry_point in pkg_resources.iter_entry_points('antibot'):
        object = entry_point.load()
        if issubclass(object, Module):
            yield object



import logging

import bottle as bottle
from injector import Injector, inject

from antibot.internal.backend.installer import PluginInstaller
from antibot.internal.module import AntibotModule
from antibot.internal.plugins import find_plugins, find_modules, PluginsCollection
from antibot.internal.scheduler import Scheduler


class Main:
    @inject
    def __init__(self, scheduler: Scheduler, plugins: PluginsCollection, installer: PluginInstaller):
        self.scheduler = scheduler
        self.plugins = plugins
        self.installer = installer

    def run(self):
        for plugin in self.plugins:
            self.installer.install_plugin(plugin)
        self.scheduler.bootstrap()
        bottle.run(port=5001, host='0.0.0.0', debug=True)


def run():
    logging.basicConfig(level=logging.INFO)

    module = AntibotModule(list(find_plugins()), list(find_modules()))
    injector = Injector(module)

    main = injector.get(Main)
    main.run()


if __name__ == '__main__':
    run()



from datetime import datetime
from json import dumps

from injector import inject

from antibot.decorators import ws, block_action, command
from antibot.internal.backend.debugger import Debugger, QueryCatcher
from antibot.plugin import AntibotPlugin
from antibot.slack.api import SlackApi
from antibot.slack.message import Message, Element


class BasePlugin(AntibotPlugin):
    @inject
    def __init__(self):
        super().__init__('Base')

    @ws('/', method='GET')
    def hello(self):
        return 'it\'s working'


DISMISS_ACTION = 'dismiss'
DISMISS_BUTTON = Element.button(DISMISS_ACTION, 'Dismiss')


class DismissActionPlugin(AntibotPlugin):
    def __init__(self):
        super().__init__('DismissAction')

    @block_action(action_id=DISMISS_ACTION)
    def on_dismiss(self):
        return Message(delete_original=True)


class DebuggerPlugin(AntibotPlugin):
    @inject
    def __init__(self, debugger: Debugger, api: SlackApi):
        super().__init__('Debug')
        self.debugger = debugger
        self.api = api

    @command('/debug')
    def catch_queries(self, params: str):
        nb_queries = int(params)

        def callback(query: dict):
            date = datetime.now().isoformat()
            self.api.upload_file('bot',
                                 filename='debug-{}-query.json'.format(date),
                                 title='Antibot debug query from {}'.format(date),
                                 content=dumps(query, indent=2).encode('utf-8'))

        self.debugger.add_hook(QueryCatcher(nb_queries, callback))



from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Union, Iterable

from pyckson import no_camel_case


class ResponseType(Enum):
    in_channel = 'in_channel'
    ephemeral = 'ephemeral'


class ActionType(Enum):
    button = 'button'
    select = 'select'


class ActionStyle(Enum):
    default = 'default'
    primary = 'primary'
    danger = 'danger'


@no_camel_case
@dataclass
class PostMessageReply:
    channel: str
    ts: str


@no_camel_case
@dataclass
class Text:
    type: str
    text: str

    @staticmethod
    def mrkdwn(text: str) -> 'Text':
        return Text('mrkdwn', text)

    @staticmethod
    def plain(text: str) -> 'Text':
        return Text('plain_text', text)


@no_camel_case
@dataclass
class Option:
    text: Text
    value: str

    @staticmethod
    def of(value: str, text: Optional[str] = None) -> 'Option':
        text = text or value
        return Option(Text.plain(text), value)


@no_camel_case
@dataclass
class OptionGroup:
    label: Text
    options: List[Option]

    @staticmethod
    def of(label: str, options: List[Option]) -> 'OptionGroup':
        return OptionGroup(Text.plain(label), options)


@no_camel_case
@dataclass
class Confirm:
    title: Text
    text: Text
    confirm: Text
    deny: Text

    @staticmethod
    def of(title: str, text: str, confirm: str, deny: str) -> 'Confirm':
        return Confirm(Text.plain(title), Text.mrkdwn(text),
                       Text.plain(confirm), Text.plain(deny))


@no_camel_case
@dataclass
class Element:
    type: str
    action_id: str
    text: Optional[Text] = None
    options: List[Option] = None
    option_groups: List[OptionGroup] = None
    initial_option: Optional[Option] = None
    initial_date: Optional[str] = None
    initial_channel: Optional[str] = None
    initial_user: Optional[str] = None
    value: Optional[str] = None
    style: Optional[ActionStyle] = None
    confirm: Optional[Confirm] = None
    placeholder: Optional[Text] = None
    initial_value: Optional[str] = None

    @staticmethod
    def button(action_id: str, text: str, style: Optional[ActionStyle] = None,
               value: Optional[str] = None, confirm: Optional[Confirm] = None) -> 'Element':
        return Element('button', action_id=action_id, text=Text.plain(text),
                       style=style, value=value, confirm=confirm)

    @staticmethod
    def select(action_id: str, placeholder: str, options: List[Option],
               initial_option: Optional[Option] = None) -> 'Element':
        return Element('static_select', action_id=action_id, placeholder=Text.plain(placeholder),
                       options=options, initial_option=initial_option)

    @staticmethod
    def group_select(action_id: str, placeholder: str, groups: List[OptionGroup],
                     initial_option: Optional[Option] = None) -> 'Element':
        return Element('static_select', action_id=action_id, placeholder=Text.plain(placeholder),
                       option_groups=groups, initial_option=initial_option)

    @staticmethod
    def select_channel(action_id: str, placeholder: str, initial_channel: Optional[str] = None):
        return Element('channels_select', action_id=action_id, placeholder=Text.plain(placeholder),
                       initial_channel=initial_channel)

    @staticmethod
    def select_user(action_id: str, placeholder: str, initial_user: Optional[str] = None):
        return Element('users_select', action_id=action_id, placeholder=Text.plain(placeholder),
                       initial_user=initial_user)

    @staticmethod
    def datepicker(action_id: str, placeholder: str, initial_date: Optional[str] = None):
        return Element('datepicker', action_id=action_id, placeholder=Text.plain(placeholder),
                       initial_date=initial_date)

    @staticmethod
    def input(action_id: str, placeholder: Optional[str] = None, initial_value: Optional[str] = None):
        return Element('plain_text_input', action_id=action_id,
                       placeholder=Text.plain(placeholder) if placeholder else None,
                       initial_value=initial_value)

    @staticmethod
    def radio(action_id: str, options: List[Option], initial_option: Optional[Option] = None):
        return Element('radio_buttons', action_id=action_id, options=options, initial_option=initial_option)


@no_camel_case
@dataclass
class Block:
    type: str
    text: Optional[Text] = None
    elements: List[Union[Text, Element]] = None
    accessory: Optional[Element] = None
    title: Optional[Text] = None
    image_url: Optional[str] = None
    alt_text: Optional[str] = None
    label: Optional[Text] = None
    element: Optional[Element] = None
    block_id: Optional[str] = None
    optional: Optional[bool] = None

    @staticmethod
    def section(text: str, accessory: Optional[Element] = None) -> 'Block':
        return Block('section', text=Text('mrkdwn', text), accessory=accessory)

    @staticmethod
    def actions(*elements: Element) -> 'Block':
        return Block('actions', elements=list(elements))

    @staticmethod
    def divider() -> 'Block':
        return Block('divider')

    @staticmethod
    def context(text: str) -> 'Block':
        return Block('context', elements=[Text.mrkdwn(text)])

    @staticmethod
    def image(title: str, url: str, alt_text: str):
        return Block('image', title=Text.plain(title), image_url=url, alt_text=alt_text)

    @staticmethod
    def input(block_id: str, label: str, element: Element, optional: bool = False) -> 'Block':
        return Block('input', block_id=block_id, label=Text.plain(label), element=element, optional=optional)


@no_camel_case
@dataclass
class View:
    type: str
    callback_id: str
    title: Text
    blocks: List[Block]
    submit: Optional[Text]
    private_metadata: Optional[str]
    notify_on_close: Optional[bool] = False

    @staticmethod
    def modal(callback_id: str, title: str, blocks: Iterable[Block], submit: Optional[str] = None,
              notify_on_close: bool = False, private_metadata: Optional[str] = None) -> 'View':
        submit = Text.plain(submit) if submit else None
        return View('modal', callback_id=callback_id, title=Text.plain(title), blocks=list(blocks), submit=submit,
                    notify_on_close=notify_on_close, private_metadata=private_metadata)


@dataclass
class ViewError:
    block_id: str
    message: str


@no_camel_case
@dataclass
class Message:
    text: Optional[str] = None
    response_type: ResponseType = ResponseType.in_channel
    replace_original: bool = False
    delete_original: bool = False
    blocks: List[Block] = None

    @staticmethod
    def ephemeral(blocks: Iterable[Block]):
        return Message('', blocks=list(blocks), response_type=ResponseType.ephemeral)

    @staticmethod
    def replace(blocks: Iterable[Block]):
        return Message('', blocks=list(blocks), replace_original=True)

    @staticmethod
    def delete():
        return Message(delete_original=True)



from dataclasses import dataclass
from typing import Type, Callable, Iterable

from injector import singleton, inject
from pyckson import parse, serialize

from antibot.internal.backend.constants import VIEW_SUBMIT_ID
from antibot.internal.backend.descriptor import find_method_by_attribute
from antibot.internal.backend.endpoint_runner import EndpointRunner
from antibot.plugin import AntibotPlugin
from antibot.repository.users import UsersRepository
from antibot.slack.api import SlackApi
from antibot.slack.callback import ViewSubmitPayload
from antibot.slack.message import View, ViewError


@dataclass
class ViewSubmitDescriptor:
    plugin: Type[AntibotPlugin]
    method: Callable
    callback_id: str


@singleton
class ViewSubmitRunner:
    @inject
    def __init__(self, endpoints: EndpointRunner, users: UsersRepository, api: SlackApi):
        self.endpoints = endpoints
        self.users = users
        self.api = api
        self.descriptors = []

    def install_plugin(self, plugin: Type[AntibotPlugin]):
        for method, callback_id in find_method_by_attribute(plugin, VIEW_SUBMIT_ID):
            self.descriptors.append(ViewSubmitDescriptor(plugin, method, callback_id))

    def find_callback(self, callback_id) -> Iterable[ViewSubmitDescriptor]:
        for descriptor in self.descriptors:
            if descriptor.callback_id == callback_id:
                yield descriptor

    def run(self, payload: dict):
        message = parse(ViewSubmitPayload, payload)
        for descriptor in self.find_callback(message.view.callback_id):
            user = self.users.get_user(message.user.id)
            reply = self.endpoints.run(descriptor.plugin, descriptor.method,
                                       user=user, callback_id=message.view.callback_id,
                                       values=message.view.state.values, view_id=message.view.id,
                                       private_metadata=message.view.private_metadata)
            if isinstance(reply, View):
                return {'response_action': 'update',
                        'view': serialize(reply)}
            if isinstance(reply, ViewError):
                return {'response_action': 'errors',
                        'errors': {
                            reply.block_id: reply.message
                        }}



import traceback
from contextlib import contextmanager
from datetime import datetime
from json import dumps
from typing import Optional, Callable

from injector import inject, singleton

from antibot.internal.configuration import Configuration
from antibot.slack.api import SlackApi


class DebugHook:
    def on_query(self, query: dict) -> bool:
        """Return False to dismiss the hook"""
        pass


class QueryCatcher(DebugHook):
    def __init__(self, count: int, callback: Callable[[dict], None]):
        self.count = count
        self.callback = callback

    def on_query(self, query: dict):
        self.callback(query)
        self.count -= 1
        if self.count == 0:
            return False
        return True


@singleton
class Debugger:
    @inject
    def __init__(self, api: SlackApi, configuration: Configuration):
        self.api = api
        self.configuration = configuration
        self.hooks = []

    @contextmanager
    def wrap(self, query: Optional[dict] = None):
        try:
            self.process_hooks(query)
            yield
        except Exception:
            if self.configuration.prod:
                date = datetime.now().isoformat()
                self.api.upload_file('bot',
                                     filename='error-{}-stacktrace.txt'.format(date),
                                     title='Antibot error stacktrace from {}'.format(date),
                                     content=traceback.format_exc().encode('utf-8'))
                if query:
                    self.api.upload_file('bot',
                                         filename='error-{}-query.json'.format(date),
                                         title='Antibot error query from {}'.format(date),
                                         content=dumps(query, indent=2).encode('utf-8'))
            raise

    def process_hooks(self, query: dict):
        for hook in self.hooks:
            res = hook.on_query(query)
            if not res:
                self.hooks.remove(hook)

    def add_hook(self, hook: DebugHook):
        self.hooks.append(hook)



from typing import Optional

from injector import singleton, inject

from antibot.slack.api import SlackApi
from antibot.user import User


@singleton
class UsersRepository:
    @inject
    def __init__(self, api: SlackApi):
        self.api = api
        users = list(self.api.list_users())
        self.users_by_id = {user.id: user for user in users}
        self.users_by_mail = {user.email: user for user in users}

    def get_user(self, user_id: str) -> Optional[User]:
        return self.users_by_id.get(user_id)

    def get_by_email(self, email: str) -> Optional[User]:
        return self.users_by_mail.get(email)



import logging
from io import BytesIO
from typing import Iterator

import requests
from injector import inject, singleton
from pyckson import parse, serialize, dumps
from requests import HTTPError
from slack import WebClient

from antibot.internal.slack.channel import Channel
from antibot.internal.slack.upload import File
from antibot.internal.slack.user import Member
from antibot.slack.message import Message, PostMessageReply, View
from antibot.user import User


@singleton
class SlackApi:
    @inject
    def __init__(self, client: WebClient):
        self.client = client

    def list_users(self) -> Iterator[User]:
        result = self.client.api_call('users.list')
        for member in result['members']:
            member = parse(Member, member)
            name = member.profile.display_name or member.profile.real_name
            yield User(member.id, name, member.profile.email)

    def get_channel(self, channel_id) -> Channel:
        result = self.client.channels_info(channel=channel_id)
        channel = parse(Channel, result['channel'])
        return channel

    def post_message(self, channel_id: str, message: Message) -> PostMessageReply:
        blocks = [serialize(block) for block in message.blocks] if message.blocks is not None else None
        result = self.client.chat_postMessage(channel=channel_id, text=message.text, blocks=blocks)
        return PostMessageReply(result['channel'], result['ts'])

    def post_ephemeral(self, channel_id: str, user_id: str, message: Message) -> PostMessageReply:
        blocks = [serialize(block) for block in message.blocks] if message.blocks is not None else None
        result = self.client.chat_postEphemeral(channel=channel_id, user=user_id, text=message.text, blocks=blocks)
        return PostMessageReply(channel_id, result['message_ts'])

    def get_permalink(self, channel_id: str, timestamp: str) -> str:
        result = self.client.chat_getPermalink(channel=channel_id, message_ts=timestamp)
        return result['permalink']

    def update_message(self, channel_id: str, timestamp: str, message: Message) -> PostMessageReply:
        blocks = [serialize(block) for block in message.blocks] if message.blocks is not None else None
        result = self.client.chat_update(channel=channel_id, ts=timestamp,
                                         text=message.text, blocks=blocks)
        return PostMessageReply(result['channel'], result['ts'])

    def respond(self, response_url: str, message: Message):
        reply = requests.post(response_url, json=serialize(message))
        try:
            reply.raise_for_status()
        except HTTPError:
            logging.getLogger(__name__).error(reply.text)
            print('##### blocks #####')
            print(dumps(message.blocks))
            raise

    def upload_file(self, channel_id: str, filename: str, title: str, content: bytes):
        result = self.client.files_upload(file=BytesIO(content), filename=filename,
                                          title=title, channels=channel_id)
        print(result.data)
        return parse(File, result.data['file'])

    def upload_and_share(self, content: bytes, filename) -> File:
        result = self.user_client.files_upload(file=BytesIO(content), filename=filename)
        result = self.user_client.files_sharedPublicURL(file=result['file']['id'])
        print(result.data)
        return parse(File, result.data['file'])

    def open_modal(self, trigger_id: str, view: View) -> str:
        result = self.client.views_open(trigger_id=trigger_id, view=serialize(view))
        return result['view']['id']

    def push_modal(self, trigger_id: str, view: View) -> str:
        result = self.client.views_push(trigger_id=trigger_id, view=serialize(view))
        return result['view']['id']

    def update_view(self, parent_view_id: str, view: View) -> str:
        result = self.client.views_update(view_id=parent_view_id, view=serialize(view))
        return result['view']['id']



from typing import Type

from bottle import request, abort
from injector import inject, Injector
from pyckson import serialize

from antibot.internal.backend.constants import WS_JSON_VALUES
from antibot.internal.backend.debugger import Debugger
from antibot.internal.configuration import Configuration
from antibot.plugin import AntibotPlugin


class WsRunner:
    @inject
    def __init__(self, injector: Injector, configuration: Configuration, debugger: Debugger):
        self.injector = injector
        self.configuration = configuration
        self.debugger = debugger

    def run_ws(self, method, plugin: Type[AntibotPlugin], **kwargs):
        request_key = request.params.get('apikey') or request.headers.get('X-Gitlab-Token')
        if self.configuration.ws_api_key != request_key:
            abort(401, 'Could not verify api key')
        ip = request.get_header('X-Forwarded-For', request.environ.get('REMOTE_ADDR'))
        instance = self.injector.get(plugin)

        with self.debugger.wrap(request.json):
            reply = method(instance, **kwargs)
            if reply is not None:
                if getattr(method, WS_JSON_VALUES, False):
                    return serialize(reply)
                return reply



from dataclasses import dataclass
from typing import List, Optional

from pyckson import no_camel_case

from antibot.slack.message import Option


@no_camel_case
@dataclass
class CallbackUser:
    id: str
    name: str


@no_camel_case
@dataclass
class CallbackChannel:
    id: str
    name: str


@no_camel_case
@dataclass
class SelectedOption:
    value: str


@no_camel_case
@dataclass
class StatePayload:
    values: dict


@no_camel_case
@dataclass
class ViewPayload:
    id: str
    callback_id: str
    state: Optional[StatePayload]
    private_metadata: Optional[str]


@no_camel_case
@dataclass
class CallbackAction:
    name: str
    selected_options: List[SelectedOption] = None


@no_camel_case
@dataclass
class Container:
    message_ts: Optional[str]


@no_camel_case
@dataclass
class BlockAction:
    action_id: str
    block_id: str
    value: Optional[str]
    selected_option: Optional[Option]
    selected_date: Optional[str]
    selected_channel: Optional[str]
    selected_user: Optional[str]


@no_camel_case
@dataclass
class BlockPayload:
    user: CallbackUser
    actions: List[BlockAction]
    trigger_id: str
    container: Container
    view: Optional[ViewPayload]
    channel: Optional[CallbackChannel]
    response_url: Optional[str]


@no_camel_case
@dataclass
class ViewClosedPayload:
    def __init__(self, user: CallbackUser, view: ViewPayload):
        self.user = user
        self.view = view


@no_camel_case
@dataclass
class ViewSubmitPayload:
    def __init__(self, user: CallbackUser, view: ViewPayload):
        self.user = user
        self.view = view



import atexit
import traceback
from inspect import getmembers
from threading import Thread
from time import sleep

import schedule
from injector import inject, Injector

from antibot.internal.backend.constants import JOB_ATTR_DAILY
from antibot.internal.plugins import PluginsCollection


class SchedulerWatch:
    def __init__(self):
        self.running = True

    def run(self):
        while self.running:
            schedule.run_pending()
            sleep(1)


def find_daily_jobs(cls):
    for name, method in getmembers(cls):
        hour = getattr(method, JOB_ATTR_DAILY, None)
        if hour is not None:
            yield method, hour


class Scheduler:
    @inject
    def __init__(self, injector: Injector, plugins: PluginsCollection, watch: SchedulerWatch):
        self.plugins = plugins
        self.injector = injector
        self.watch = watch
        self.watch_thread = Thread(target=watch.run)

    def bootstrap(self):
        for plugin in self.plugins.plugins:
            for method, hour in find_daily_jobs(plugin):
                print(method, hour)
                schedule.every().day.at(hour).do(self.run, plugin, method)
        self.watch_thread.start()

        def stop():
            print('stop')
            self.watch.running = False
            self.watch_thread.join(1)

        atexit.register(lambda: stop())

    def run(self, cls, method):
        try:
            instance = self.injector.get(cls)
            method(instance)
        except Exception as e:
            traceback.print_exc()



from dataclasses import dataclass
from inspect import getmembers
from typing import Iterator, Optional, Callable, Iterable, Any, Tuple

from antibot.internal.backend.constants import CMD_ATTR, WS_ATTR


@dataclass
class CommandDescriptor:
    route: str
    method: Callable


def find_commands(cls) -> Iterator[CommandDescriptor]:
    for name, method in getmembers(cls):
        if hasattr(method, CMD_ATTR):
            yield getattr(method, CMD_ATTR)


@dataclass
class BlockActionOptions:
    block_id: Optional[str]
    action_id: Optional[str]


def find_method_by_attribute(cls, attr) -> Iterable[Tuple[Callable, Any]]:
    for name, method in getmembers(cls):
        if hasattr(method, attr):
            yield (method, getattr(method, attr))


@dataclass
class WsDescriptor:
    route: str
    http_method: str
    method: Callable


def find_ws(cls) -> Iterator[WsDescriptor]:
    for name, method in getmembers(cls):
        if hasattr(method, WS_ATTR):
            yield getattr(method, WS_ATTR)
